from dataclasses import dataclass, field
from uuid import uuid4

from frater.client import API
from frater.component import ComponentConfig
from frater.config import Config
from frater.utilities import NetworkConfig

__all__ = ['ComponentClient', 'ComponentClientConfig', 'get_component_client_config']


@dataclass
class ComponentClientConfig(Config):
    name: str = 'component_client_config'
    component_id: str = field(default_factory=lambda: str(uuid4()))
    host: str = 'localhost'
    port: int = 3000


class ComponentClient:
    def __init__(self, config: ComponentClientConfig):
        self.config = config
        self.component_api = API(config.host, config.port)

    def start(self):
        return self.component_api.get('start')

    def started(self):
        return self.component_api.get('started')

    def stop(self):
        return self.component_api.get('stop')

    def stopped(self):
        return self.component_api.get('stopped')

    def toggle_pause(self):
        return self.component_api.get('toggle_pause')

    def pause(self):
        return self.component_api.get('pause')

    def unpause(self):
        return self.component_api.get('unpause')

    def paused(self):
        return self.component_api.get('paused')

    def active(self):
        return self.component_api.get('active')

    def reset(self):
        return self.component_api.get('reset')

    def error(self):
        return self.component_api.get('error')

    def update_config(self, config):
        return self.component_api.patch('config', config)

    def get_config(self):
        return self.component_api.get('config')

    def get_state(self):
        return self.component_api.get('state')

    def get_status(self):
        return self.component_api.get('status')

    def get_endpoints(self):
        return self.component_api.get('endpoints')


def get_component_client_config(component_config: ComponentConfig, network_config: NetworkConfig):
    return ComponentClientConfig(component_id=component_config.component_id, name=component_config.name,
                                 host=network_config.host, port=network_config.port)



import copy
import functools

from .error import ValidationError


def validate_json(_func=None, default=None, completion=True):
    def decorator_validate(func):
        @functools.wraps(func)
        def wrapper_validate(data):
            if completion:
                data = add_defaults(copy.deepcopy(data), default)
            else:
                validate_helper(data, default)
            return func(data)

        return wrapper_validate

    if _func is None:
        return decorator_validate
    else:
        return decorator_validate(_func)


def validate_helper(data, default):
    for key, value in default.items():
        if key not in data:
            raise ValidationError(
                'Key {} not found in {} Object'.format(key, type(data)))
        elif type(value) != type(data[key]):
            raise ValidationError(
                'Key {} is invalid type: default - {}, data - {}'.format(
                    key, type(value), type(data[key])
                )
            )


def add_defaults(data, default):
    default_keys = set(default.keys())
    data_keys = set(data.keys())
    default_keys.difference_update(data_keys)

    for key in default_keys:
        data[key] = default[key]

    return data



from dataclasses import dataclass, field
from typing import List, Iterable

from .component import ComponentState
from .io_component import IOComponent, IOComponentConfig
from ...stream import InputStream, OutputStream

__all__ = ['BatchComponent', 'BatchComponentConfig', 'BatchComponentState']


@dataclass
class BatchComponentConfig(IOComponentConfig):
    name: str = 'batch_component_config'
    batch_size: int = 1


@dataclass
class BatchComponentState(ComponentState):
    batch: List = field(default_factory=list)


class BatchComponent(IOComponent):
    def __init__(self, config: BatchComponentConfig, input_stream: InputStream, output_stream: OutputStream):
        super(BatchComponent, self).__init__(config, input_stream, output_stream)

    def init_state(self) -> BatchComponentState:
        return BatchComponentState()

    @property
    def current_batch(self):
        return self.state.batch

    @property
    def batch_size(self):
        return self.config.batch_size

    def preprocess(self, data):
        self.add_to_batch(data)
        if self.batch_is_ready():
            return self.current_batch
        else:
            return []

    def process(self, batch: List) -> Iterable:
        raise NotImplementedError

    def after_process(self, data):
        if self.batch_is_ready():
            self.reset_batch()

    def on_end_of_stream(self, data):
        if len(self.current_batch) > 0:
            self.batch_lifecycle(self.current_batch)
            self.reset_batch()

    def add_to_batch(self, data):
        self.current_batch.append(data)

    def batch_is_ready(self):
        return len(self.current_batch) == self.batch_size

    def reset_batch(self):
        self.state.batch = list()

    def send_output(self, outputs):
        for output in outputs:
            self.output_stream.send(output)

    def batch_lifecycle(self, batch):
        output = self.process(batch)
        self.after_process(output)
        postprocessed_output = self.postprocess(output)
        self.after_postprocess(postprocessed_output)
        self.send_output(postprocessed_output)



from enum import IntEnum
from typing import Dict

from ..io.deserializers import json_deserializers
from ..io.serializers import json_serializers
from ..validation.json import validate_json


class StreamState(IntEnum):
    STOP = 0
    START = 1
    END = 2  # end of sequence


STREAM_STATE_JSON_DEFAULTS = {
    'data_type': 'stream_state',
    'stream_state': 0
}


@json_deserializers.register('stream_state')
@validate_json(default=STREAM_STATE_JSON_DEFAULTS, completion=True)
def json_to_stream_state(data: Dict):
    return StreamState(data['stream_state'])


@json_serializers.register(StreamState)
def stream_state_to_json(state: StreamState):
    return {'stream_state': state.value, 'data_type': 'stream_state'}


def is_end_of_stream(data):
    return isinstance(data, StreamState) and data == StreamState.END


def is_start_of_stream(data):
    return isinstance(data, StreamState) and data == StreamState.START



import logging
import sys
from dataclasses import dataclass
from enum import IntEnum

from .config import HandlerConfig
from .factory import handler_factory, handler_config_factory

__all__ = ['Standard', 'StandardHandlerConfig', 'StandardHandler']


class Standard(IntEnum):
    OUT = 0
    ERR = 1


@handler_config_factory.register('std')
@dataclass
class StandardHandlerConfig(HandlerConfig):
    stream: Standard = Standard.OUT


@handler_factory.register('std')
class StandardHandler(logging.StreamHandler):
    def __init__(self, config: StandardHandlerConfig):
        stream = self.get_stream(config.stream)
        super(StandardHandler, self).__init__(stream)

    @staticmethod
    def get_stream(stream: Standard):
        if stream == Standard.OUT:
            return sys.stdout
        else:
            return sys.stderr



from dataclasses import dataclass, field
from typing import Callable, List, Dict

from .middleware import get_default_middleware
from .server_manager import ServerManager
from ..client import SystemClient, SystemClientConfig, get_component_client_config
from ..component import ComponentManager, ComponentBuilder
from ..config import Config
from ..stream import StreamFactoryConfig, get_streams
from ..utilities import NetworkConfig

__all__ = ['ComponentServerConfig', 'serve_component']


@dataclass
class ComponentServerConfig(Config):
    name: str = 'component_server_config'
    component_config: Dict = field(default_factory=dict)
    stream_config: StreamFactoryConfig = field(default_factory=StreamFactoryConfig)
    system_config: SystemClientConfig = field(default_factory=SystemClientConfig)
    network_config: NetworkConfig = field(default_factory=NetworkConfig)


def serve_component(component_class: type, component_config_class: type,
                    server_config: ComponentServerConfig, middleware: List[Callable] = None):
    # connect to system manager
    system_client = SystemClient(server_config.system_config)
    # check dependencies for streams
    system_client.wait_for_dependencies(server_config.stream_config.dependencies)
    # create streams
    input_stream, output_stream = get_streams(server_config.stream_config)
    # check dependencies for component
    component_config = component_config_class.from_dict(server_config.component_config)
    system_client.wait_for_dependencies(component_config.dependencies)
    # create component
    component = ComponentBuilder.build(component_class, component_config, input_stream, output_stream)
    # create component manager
    component_manager = ComponentManager(component)
    # setup server
    if not middleware:
        middleware = get_default_middleware()
    server_manager = ServerManager(middleware)
    # register component endpoints with server
    server_manager.register_endpoints(component_manager.get_all_handlers())
    # register component with system
    client_config = get_component_client_config(server_config.component_config, server_config.network_config)
    system_client.register_component(client_config)
    return server_manager.get_server()



import json
from typing import Callable

__all__ = ['json_deserializers', 'json_to_frater', 'get_kafka_deserializer']

from ..data_type import data_types
from ..factory import Factory

json_deserializers = Factory()


def json_to_frater(data):
    d_type = data['data_type'] if 'data_type' in data else ''
    if isinstance(data, list):
        return [json_to_frater(item) for item in data]
    elif d_type in data_types:
        return data_types[d_type].from_dict(data)
    elif d_type in json_deserializers:
        return json_deserializers[d_type](data)
    else:
        return data


def get_kafka_deserializer() -> Callable:
    return lambda m: json_to_frater(json.loads(m.decode('utf-8')))



import json
from dataclasses import dataclass

from .factory import *
from .stream import InputStream, OutputStream, StreamConfig
from ..io import frater_to_json, json_to_frater

__all__ = ['JSONStreamConfig', 'JSONInputStream', 'JSONOutputStream']


@input_stream_configs.register('json')
@output_stream_configs.register('json')
@dataclass
class JSONStreamConfig(StreamConfig):
    filename: str = ''


@input_stream_factory.register('json')
class JSONInputStream(InputStream):
    def __init__(self, config: JSONStreamConfig):
        super(JSONInputStream, self).__init__(config)
        self.input_file = self.open_file()

    def __iter__(self):
        for line in self.input_file:
            yield json_to_frater(json.loads(line.strip()))

    def close(self):
        self.input_file.close()

    def open(self):
        self.input_file = self.open_file()

    def open_file(self):
        return open(self.config.filename, 'r')


@output_stream_factory.register('json')
class JSONOutputStream(OutputStream):
    def __init__(self, config: JSONStreamConfig, data_type: type = None):
        super(JSONOutputStream, self).__init__(config)
        self.output_file = self.open_file()

    def send(self, data):
        json_output = self.get_json_output(data)
        self.output_file.write(json_output)

    def close(self):
        self.output_file.close()

    def open(self):
        self.output_file = self.open_file()

    def open_file(self):
        return open(self.config.filename, 'a')

    @staticmethod
    def get_json_output(data):
        return json.dumps(frater_to_json(data)) + '\n'



from dataclasses import dataclass
from typing import List

from ..config import Config


@dataclass
class StreamConfig(Config):
    data_type: str = ''


class Stream:
    """
    Base Stream class. Only inherited by InputStream and Output Stream
    """

    def __init__(self, config: StreamConfig = None):
        self.config = config

    @property
    def data_type(self):
        return self.config.data_type

    def close(self):
        pass


class InputStream(Stream):
    def __init__(self, data_type=None):
        super(InputStream, self).__init__(data_type)

    def __iter__(self):
        raise NotImplemented


class OutputStream(Stream):
    def send(self, data):
        raise NotImplemented

    def __call__(self, *args, **kwargs):
        return self.send(*args, **kwargs)


class MultiOutputStream(OutputStream):
    """
    MultiOutputStream allows for multiple output sources
    """

    def __init__(self, output_streams: List[OutputStream] = None, data_type: type = None):
        super(MultiOutputStream, self).__init__(data_type)
        if output_streams is None:
            output_streams: List[OutputStream] = list()
        self.output_streams = output_streams

    def send(self, data):
        for output_stream in self.output_streams:
            output_stream.send(data)

    def add_output_stream(self, output_stream: OutputStream):
        self.output_streams.append(output_stream)



from dataclasses import dataclass, field

from pymongo import MongoClient

from .factory import *
from .stream import InputStream, OutputStream, StreamConfig
from ..io import json_to_frater, frater_to_json


@output_stream_configs.register('mongo')
@input_stream_configs.register('mongo')
@dataclass
class MongoStreamConfig(StreamConfig):
    name: str = 'mongo_stream_config'
    host: str = 'localhost'
    port: int = 27017
    db: str = 'db'
    collection: str = 'collection'
    filter: dict = field(default_factory=dict)


@input_stream_factory.register('mongo')
class MongoInputStream(InputStream):
    def __init__(self, config: MongoStreamConfig):
        super(MongoInputStream, self).__init__(config)
        self.client = MongoClient(self.config.host, self.config.port)

    @property
    def db(self):
        return self.client[self.config.db]

    @property
    def collection(self):
        return self.db[self.config.collection]

    def __iter__(self):
        for item in self.collection.find(filter=self.config.filter):
            yield json_to_frater(item)


@output_stream_factory.register('mongo')
class MongoOutputStream(OutputStream):
    def __init__(self, config: MongoStreamConfig):
        super(MongoOutputStream, self).__init__(config)
        self.client = MongoClient(self.config.host, self.config.port)

    @property
    def db(self):
        return self.client[self.config.db]

    @property
    def collection(self):
        return self.db[self.config.collection]

    def send(self, data):
        self.collection.insert_one(frater_to_json(data))



from .api import API
from .component import *
from .system import *

__all__ = ['ComponentClient', 'ComponentClientConfig', 'get_component_client_config', 'SystemClient',
           'SystemClientConfig',
           'API']



from io import BytesIO

import requests
from PIL import Image


def load_image_from_file(filename):
    return Image.open(filename)


def save_image_to_file(image: Image.Image, filename):
    image.save(filename)


def load_image_from_url(url):
    response = requests.get(url)
    return Image.open(BytesIO(response.content))



from dataclasses import dataclass

from frater.config import Config


@dataclass
class NetworkConfig(Config):
    name: str = 'network_config'
    host: str = '0.0.0.0'
    port: int = 3000
    protocol: str = 'http://'

    @property
    def uri(self) -> str:
        return f'{self.protocol}{self.host}:{self.port}'



import logging
import time

from kafka import KafkaClient

logger = logging.getLogger()


def wait_for_kafka_servers(servers, max_retries=10, sleep=1):
    logger.info(f'waiting for servers: {servers}')
    for _ in range(max_retries):
        available = kafka_servers_available(servers)
        logging.info(f'Hosts: {", ".join(servers)}, Available: {available}')
        if available:
            break
        time.sleep(sleep)
    else:
        raise ConnectionError(f'Failed to connect to Kafka Servers {servers}. Attempted max retries.')


def kafka_servers_available(servers):
    try:
        client = KafkaClient(bootstrap_servers=servers)
        client.close()
    except Exception:
        return False
    return True


def end_of_sequence(headers):
    return bool(headers['EOS'])



from dataclasses import dataclass, field
from logging import Logger

from .stream import OutputStream, StreamConfig
from ..logging import summarize
from ..logging.logger import LoggerConfig


@dataclass
class LoggerOutputStreamConfig(StreamConfig):
    logger_config: LoggerConfig = field(default_factory=LoggerConfig)


class LoggerOutputStream(OutputStream):
    def __init__(self, logger: Logger, multiline=True, data_type: type = None):
        super(LoggerOutputStream, self).__init__(data_type)
        self.logger = logger
        self.multiline = multiline

    def send(self, data):
        self.logger.info(summarize(data, self.multiline))



from dataclasses import dataclass, field
from typing import List

from dataclasses_json import DataClassJsonMixin

from frater.category import Category


@dataclass
class Dataset(DataClassJsonMixin):
    name: str = ''
    labels: List[str] = field(default_factory=list)

    def __getitem__(self, item: int):
        return self.get_category(item)

    def __len__(self):
        return len(self.labels)

    def get_category(self, index):
        return Category(index, self.labels[index], self.name)

    def get_category_by_label(self, label):
        index = self.labels.index(label)
        return self.get_category(index)



from dataclasses import dataclass, field
from typing import Callable, List

from dataclasses_json import DataClassJsonMixin


@dataclass
class Handler(DataClassJsonMixin):
    """
    :param endpoint: API endpoint for handler func
    :param name: Flask view function name
    :param func: handler function
    :param methods: HTTP methods for handler function
    """
    endpoint: str
    name: str
    func: Callable
    methods: List[str] = field(default_factory=lambda: ['GET'])

    def to_dict(self, **kwargs):
        return {'endpoint': self.endpoint, 'name': self.name, 'methods': self.methods}



from .builder import *
from .component import *
from .manager import *

__all__ = ['Component', 'ComponentConfig', 'ComponentState', 'IOComponent', 'IOComponentConfig', 'InputComponent',
           'InputComponentConfig', 'OutputComponent', 'OutputComponentConfig', 'BatchComponent', 'BatchComponentConfig',
           'ComponentBuilder', 'ComponentManager']



from dataclasses import dataclass, field
from logging import Handler
from typing import List

from kafka import KafkaProducer

from .config import HandlerConfig
from .factory import handler_factory, handler_config_factory
from ...io import get_kafka_serializer

__all__ = ['KafkaHandler', 'KafkaHandlerConfig']


@handler_config_factory.register('kafka')
@dataclass
class KafkaHandlerConfig(HandlerConfig):
    name: str = 'kafka_handler_config'
    topic: str = 'logging'
    servers: List[str] = field(default_factory=lambda: ['localhost:9092'])


@handler_factory.register('kafka')
class KafkaHandler(Handler):
    def __init__(self, config: KafkaHandlerConfig):
        """Initialize an instance of the kafka handler."""
        super(KafkaHandler, self).__init__(level=config.level)
        self.producer = KafkaProducer(bootstrap_servers=config.servers, value_serializer=get_kafka_serializer(),
                                      linger_ms=10)
        self.topic = config.topic

    def emit(self, record):
        """Emit the provided record to the kafka_client producer."""
        # drop kafka logging to avoid infinite recursion
        if 'kafka.' in record.name:
            return

        try:
            # apply the logger formatter
            msg = self.format(record)
            self.producer.send(self.topic, {'message': msg})
            self.flush(timeout=1.0)
        except Exception as e:
            Handler.handleError(self, record)

    def flush(self, timeout=None):
        """Flush the objects."""
        self.producer.flush(timeout=timeout)

    def close(self):
        """Close the producer and clean up."""
        self.acquire()
        try:
            if self.producer:
                self.producer.close()

            Handler.close(self)
        finally:
            self.release()



import json
from dataclasses import dataclass, field
from typing import Dict
from uuid import uuid4

import yaml
from dataclasses_json import DataClassJsonMixin


@dataclass
class Config(DataClassJsonMixin):
    """Base class for building configs in Frater. Used in :py:mod:`frater.component`, :py:mod:`frater.stream` and \
    several other modules


     :param name: name for the given config. Defaults to "config"
     :type name: str
     :param config_id: config_id for the given config. Generally should let this stay default,\
     unless deserializing an existing config. Will default to a random uuid4
     :type config_id: str
     :param meta: dictionary for storing extra metadata passed into config that is not an existing field for the\
     class. If using the default constructor, then you just supply the dictionary. If using any of the constructors\
     from json/yaml/dict, it will gather extra parameters for those fields.
     :type meta: dict
    """
    name: str = 'config'
    config_id: str = field(default_factory=lambda: str(uuid4()))
    meta: Dict = field(default_factory=dict)

    def update_from_dict(self, d: Dict):
        for k, v in d.items():
            if k == 'meta':
                self.meta.update(v)
            if hasattr(self, k):
                k_class = type(getattr(self, k))
                if issubclass(k_class, Config):
                    v = k_class.from_dict(v)
                elif issubclass(k_class, DataClassJsonMixin):
                    v = k_class.from_dict(v)
                setattr(self, k, v)
            else:
                self.meta[k] = v

    def update_from_json(self, s: str):
        d = json.loads(s)
        self.update_from_dict(d)

    def update_from_yaml(self, s: str):
        d = yaml.load(s, yaml.FullLoader)
        self.update_from_dict(d)

    @classmethod
    def from_dict(cls, config: Dict, **kwargs):
        required_config, optional_config = cls.separate_config(config)
        return cls(**required_config, meta=optional_config)

    @classmethod
    def separate_config(cls, config: Dict):
        required_config = dict()
        optional_config = dict()

        annotations = get_config_annotations(cls)
        for key, value in config.items():
            if key == 'meta':
                optional_config.update(value)
            elif key in annotations:
                required_config[key] = process_config_value(annotations[key], value)
            else:
                optional_config[key] = value

        return required_config, optional_config

    @classmethod
    def from_yaml(cls, s: str, from_file=False, **kwargs):
        if from_file:
            with open(s) as f:
                d = yaml.load(f, yaml.FullLoader)
                return cls.from_dict(d, **kwargs)

        return cls.from_dict(yaml.load(s), **kwargs)

    def to_yaml(self, filename=''):
        d = self.to_dict()
        if filename:
            with open(filename, 'w') as f:
                return yaml.dump(d, f)

        return yaml.dump(d)


def process_config_value(annotation_type: type, value):
    if isinstance(value, list):
        list_type = getattr(annotation_type, '__args__')[0]
        return [process_config_value(list_type, item) for item in value]
    elif annotation_type is Dict:
        return dict(value)
    elif issubclass(annotation_type, Config):
        return annotation_type.from_dict(value)
    else:
        return value


def get_config_annotations(cls: type) -> Dict:
    annotations = dict()
    if issubclass(cls, Config) and cls is not Config:
        for super_class in cls.mro()[1:-1]:
            if super_class is DataClassJsonMixin:
                break
            annotations.update(get_config_annotations(super_class))

    annotations.update(cls.__annotations__)

    return annotations



from typing import List

from flask import Flask

from ..utilities import Handler

__all__ = ['ServerManager']


class ServerManager:
    def __init__(self, middleware):
        self.server = Flask(__name__)
        self.middleware = middleware

    def get_server(self):
        return self.server

    def add_handler(self, handler: Handler):
        self.server.add_url_rule(handler.endpoint, handler.name, self.middleware(handler.func), methods=handler.methods)

    def register_endpoints(self, handlers: List[Handler]):
        for handler in handlers:
            self.add_handler(handler)



from ..component import IOComponent, ComponentConfig, OutputComponent, InputComponent, Component

from ...stream import OutputStream, InputStream


class ComponentBuildError(BaseException):
    pass


class ComponentBuilder:
    @staticmethod
    def build(component_class: type, component_config: ComponentConfig,
              input_stream: InputStream = None, output_stream: OutputStream = None) -> Component:
        if issubclass(component_class, IOComponent):
            if input_stream is None or output_stream is None:
                raise ComponentBuildError(f'InputStream and OutputStream may not be None for IOComponent')
            return component_class(component_config, input_stream, output_stream)
        elif issubclass(component_class, OutputComponent):
            if output_stream is None:
                raise ComponentBuildError(f'OutputStream may not be None for OutputComponent')
            return component_class(component_config, output_stream)
        elif issubclass(component_class, InputComponent):
            if input_stream is None:
                raise ComponentBuildError(f'InputStream may not be None for InputComponent')
            return component_class(component_config, input_stream)
        else:
            raise ComponentBuildError(f'Type {component_class} is not a valid component')



def get_default_summary(data, eol=' '):
    return str(data) + eol


def get_summary(data, summarizer=None, multiline=True):
    if summarizer is None:
        summarizer = get_default_summary
    eol = '\n' if multiline else ' '
    return summarizer(data, eol)


def summarize(data, multiline=True):
    return data.summary(multiline) if hasattr(data, 'summary') else get_summary(data, multiline=multiline)



from functools import reduce
from typing import Callable

from flask import request, jsonify

from ..utilities.function import has_argument

__all__ = ['reduce_middleware', 'get_default_middleware', 'response_middleware', 'identity_middleware']


def reduce_middleware(middleware):
    def wrapper_func(handler):
        if len(middleware) == 0:
            return identity_middleware(handler)
        elif len(middleware) == 1:
            return middleware[0](handler)
        else:
            return reduce(lambda f, g: g(f), middleware, handler)

    return wrapper_func


def get_default_middleware():
    return response_middleware


def response_middleware(handler: Callable):
    def wrapper():
        if request.method in {'POST', 'PATCH', 'PUT', 'DELETE'}:
            return jsonify(handler(request.json))
        else:
            if has_argument(handler, 'params'):
                return jsonify(handler(request.args.to_dict()))
            return jsonify(handler())

    return wrapper


def identity_middleware(handler):
    return handler



from setuptools import setup, find_packages

packages = find_packages(exclude=['tests*'])

with open('requirements.txt') as f:
    dependencies = list(line.strip() for line in f.readlines())

with open('README.md') as f:
    description = f.read()

setup(
    name='frater',
    version='0.3.1.9',
    packages=packages,
    license='MIT',
    url='https://github.com/frater-sdk/frater',
    author='John Henning',
    description='A Machine Learning and Data-Driven Systems Framework and Toolkit',
    long_description=description,
    long_description_content_type='text/markdown',
    author_email='john.l.henning@ibm.com',
    install_requires=dependencies,
    python_requires='>=3.6'
)



from .build import get_streams
from .factory import *
from .json import JSONInputStream, JSONOutputStream
from .kafka import KafkaInputStream, KafkaOutputStream
from .log import LoggerOutputStream
from .mongo import MongoInputStream, MongoOutputStream, MongoStreamConfig
from .stream import InputStream, OutputStream, MultiOutputStream, StreamConfig
from .stream_state import StreamState, is_start_of_stream, is_end_of_stream

__all__ = [
    'KafkaInputStream', 'KafkaOutputStream', 'MongoInputStream', 'MongoOutputStream', 'MongoStreamConfig',
    'InputStream', 'OutputStream', 'MultiOutputStream', 'LoggerOutputStream', 'StreamFactoryConfig',
    'StreamState', 'is_end_of_stream', 'is_start_of_stream', 'InputStreamFactoryConfig', 'OutputStreamFactoryConfig',
    'get_streams', 'input_stream_factory', 'output_stream_factory', 'input_stream_configs', 'output_stream_configs',
    'StreamConfig'
]
